use super::Member;

use once_cell::sync::Lazy;
use std::collections::{BTreeSet, HashMap, HashSet};

// Create a store for saving information when migrating to a new members list
static MEMBERS_STORE: Lazy<MembersStore> = Lazy::new(|| MembersStore::new());

struct MembersStore {
    store: HashMap<u16, Vec<Member>>,
    count: u16,
}

impl MembersStore {
    fn new() -> Self {
        Self {
            store: HashMap::new(),
            count: 0,
        }
    }

    fn insert(&mut self, input: Vec<Member>) -> u16 {
        let count = self.count + 1;
        self.store.insert(count, input);
        count
    }

    fn get(&self, key: &u16) -> Option<&Vec<Member>> {
        self.store.get(key)
    }

    fn remove(&mut self, key: &u16) -> Option<Vec<Member>> {
        self.store.remove(key)
    }
}

// Create a row for the csv file
#[derive(Debug, serde::Deserialize)]
struct Row {
    #[serde(rename = "Relatiecode")]
    id: String,
    #[serde(rename = "Roepnaam")]
    first_name: String,
    #[serde(rename = "Tussenvoegsel(s)")]
    middle_name: String,
    #[serde(rename = "Achternaam")]
    last_name: String,
    #[serde(rename = "E-mail")]
    email: String,
    #[serde(rename = "Verenigingssporten")]
    hours: String,
    #[serde(rename = "Diploma dropdown 1")]
    diploma: Option<String>,
}

impl Row {
    fn to_member(&self) -> Member {
        Member {
            id: self.id.trim().to_string(),
            name: super::Name {
                first: self.first_name.clone(),
                full: self.generate_full_name(),
            },
            hours: self.generate_hours(),
            groups: BTreeSet::new(),
            diploma: self.diploma.clone(),
            registration_token: None,
        }
    }

    // Get the hour data from the raw string
    fn generate_hours(&self) -> BTreeSet<String> {
        let mut hours: BTreeSet<String> = BTreeSet::new();

        let group_parts: Vec<&str> = self.hours.split(", ").collect();

        for group in group_parts {
            let hour_parts: Vec<&str> = group.split(" - ").collect();

            for part in hour_parts {
                if &*part != "Groep" {
                    hours.insert(part.to_string());
                }
            }
        }

        hours
    }

    // Generate the full name from 3 parts
    fn generate_full_name(&self) -> String {
        let mut parts: Vec<&str> = vec![];

        parts.push(&self.first_name);
        parts.push(&self.middle_name);
        parts.push(&self.last_name);

        let filtered_strings: Vec<&str> = parts.into_iter().filter(|s| !s.is_empty()).collect();

        filtered_strings.join(" ")
    }
}

// Convert the raw csv file to rust objects
fn csv_to_rows(input: String) -> Result<Vec<Row>, Box<dyn std::error::Error>> {
    let mut members: Vec<Row> = vec![];

    let mut rdr = csv::Reader::from_reader(input.as_bytes());

    for result in rdr.deserialize() {
        let row: Row = result?;
        members.push(row);
    }

    Ok(members)
}

// Covert the rows to formatted members
fn rows_to_members(rows: Vec<Row>) -> Vec<Member> {
    let mut members: Vec<Member> = vec![];

    for row in rows {
        members.push(row.to_member());
    }

    members
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn csv_to_members_test() -> Result<(), String> {
        let data = "Relatiecode,Volledige naam (1),Roepnaam,Tussenvoegsel(s),Achternaam,E-mail,2e E-mail,Verenigingssporten,Diploma dropdown 1
            D000001,\"Last, First\",First,,Last,first.last@example.com,first.last@example.nl,\"Groep - Wedstrijd - Zaterdag, Groep - Z5 - Zaterdag\",LS1
            D000002,\"Last2, First2\",First2,,Last2,first1.last@example.nl,,Groep - Z5 - Zaterdag,ZR4".to_string();

        let expected = vec![
            Member {
                id: "D000001".to_string(),
                name: super::super::Name {
                    first: "First".to_string(),
                    full: "First Last".to_string(),
                },
                hours: BTreeSet::from([
                    "Wedstrijd".to_string(),
                    "Z5".to_string(),
                    "Zaterdag".to_string(),
                ]),
                groups: BTreeSet::new(),
                diploma: Some("LS1".to_string()),
                registration_token: None,
            },
            Member {
                id: "D000002".to_string(),
                name: super::super::Name {
                    first: "First2".to_string(),
                    full: "First2 Last2".to_string(),
                },
                hours: BTreeSet::from(["Z5".to_string(), "Zaterdag".to_string()]),
                groups: BTreeSet::new(),
                diploma: Some("ZR4".to_string()),
                registration_token: None,
            },
        ];

        let rows = match csv_to_rows(data) {
            Ok(r) => r,
            Err(err) => return Err(err.to_string()),
        };

        let members = rows_to_members(rows);

        assert_eq!(expected, members);

        Ok(())
    }
}
